@@ -1,9 +1,7 @@
-# -*- coding: utf-8 -*-
 # Author: Franziska Horn <cod3licious@gmail.com>
 # License: MIT
 
-from __future__ import unicode_literals, division, print_function, absolute_import
-from builtins import range
+from __future__ import annotations
 import warnings
 import numba as nb
 import numpy as np
@@ -19,7 +17,7 @@ from .feateng import engineer_features, n_cols_generated, colnames2symbols
 from .featsel import select_features
 
 
-def _parse_units(units, ureg=None, verbose=0):
+def _parse_units(units: dict, ureg: pint.UnitRegistry | None = None, verbose: int = 0):
     """
     Convert a dict with string units to pint quantities.
 
@@ -40,28 +38,27 @@ def _parse_units(units, ureg=None, verbose=0):
                 parsed_units[c] = ureg.parse_expression(units[c])
             except pint.UndefinedUnitError:
                 if verbose > 0:
-                    print("[AutoFeat] WARNING: unit %r of column %r was not recognized and will be ignored!" % (units[c], c))
+                    print(f"[AutoFeat] WARNING: unit {units[c]} of column {c} was not recognized and will be ignored!")
                 parsed_units[c] = ureg.parse_expression("")
-            parsed_units[c].__dict__["_magnitude"] = 1.
+            parsed_units[c].__dict__["_magnitude"] = 1.0
     return parsed_units
 
 
 class AutoFeatModel(BaseEstimator):
-
     def __init__(
         self,
-        problem_type="regression",
-        categorical_cols=None,
-        feateng_cols=None,
-        units=None,
-        feateng_steps=2,
-        featsel_runs=5,
-        max_gb=None,
-        transformations=("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
-        apply_pi_theorem=True,
-        always_return_numpy=False,
-        n_jobs=1,
-        verbose=0,
+        problem_type: str = "regression",
+        categorical_cols: list | None = None,
+        feateng_cols: list | None = None,
+        units: dict | None = None,
+        feateng_steps: int = 2,
+        featsel_runs: int = 5,
+        max_gb: int | None = None,
+        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
+        apply_pi_theorem: bool = True,
+        always_return_numpy: bool = False,
+        n_jobs: int = 1,
+        verbose: int = 0,
     ):
         """
         multi-step feature engineering and cross-validated feature selection to generate promising additional
@@ -127,7 +124,7 @@ class AutoFeatModel(BaseEstimator):
         """
         return {k: self.__dict__[k] if k != "feature_functions_" else {} for k in self.__dict__}
 
-    def _transform_categorical_cols(self, df):
+    def _transform_categorical_cols(self, df: pd.DataFrame) -> pd.DataFrame:
         """
         Transform categorical features into 0/1 encoding.
 
@@ -138,19 +135,19 @@ class AutoFeatModel(BaseEstimator):
         """
         self.categorical_cols_map_ = {}
         if self.categorical_cols:
-            e = OneHotEncoder(sparse=False, categories="auto")
+            e = OneHotEncoder(sparse_output=False, categories="auto")
             for c in self.categorical_cols:
                 if c not in df.columns:
-                    raise ValueError("[AutoFeat] categorical_col %r not in df.columns" % c)
+                    raise ValueError(f"[AutoFeat] categorical_col {c} not in df.columns")
                 ohe = e.fit_transform(df[c].to_numpy()[:, None])
-                new_cat_cols = ["cat_%s_%r" % (str(c), i) for i in e.categories_[0]]
+                new_cat_cols = [f"cat_{c}_{i}" for i in e.categories_[0]]
                 self.categorical_cols_map_[c] = new_cat_cols
                 df = df.join(pd.DataFrame(ohe, columns=new_cat_cols, index=df.index))
             # remove the categorical column from our columns to consider
             df.drop(columns=self.categorical_cols, inplace=True)
         return df
 
-    def _apply_pi_theorem(self, df):
+    def _apply_pi_theorem(self, df: pd.DataFrame) -> pd.DataFrame:
         if self.apply_pi_theorem and self.units:
             ureg = pint.UnitRegistry(auto_reduce_dimensions=True, autoconvert_offset_to_baseunit=True)
             parsed_units = _parse_units(self.units, ureg, self.verbose)
@@ -161,18 +158,18 @@ class AutoFeatModel(BaseEstimator):
             pi_theorem_results = ureg.pi_theorem(parsed_units)
             for i, r in enumerate(pi_theorem_results, 1):
                 if self.verbose:
-                    print("[AutoFeat] Pi Theorem %i: " % i, pint.formatter(r.items()))
+                    print(f"[AutoFeat] Pi Theorem {i}: ", pint.formatter(r.items()))
                 # compute the final result by multiplying and taking the power of
                 cols = sorted(r)
                 # only use data points where non of the affected columns are NaNs
                 not_na_idx = df[cols].notna().all(axis=1)
-                ptr = df[cols[0]].to_numpy()[not_na_idx]**r[cols[0]]
+                ptr = df[cols[0]].to_numpy()[not_na_idx] ** r[cols[0]]
                 for c in cols[1:]:
-                    ptr *= df[c].to_numpy()[not_na_idx]**r[c]
-                df.loc[not_na_idx, "PT%i_%s" % (i, pint.formatter(r.items()).replace(" ", ""))] = ptr
+                    ptr *= df[c].to_numpy()[not_na_idx] ** r[c]
+                df.loc[not_na_idx, f"PT{i}_{pint.formatter(r.items()).replace(' ', '')}"] = ptr
         return df
 
-    def _generate_features(self, df, new_feat_cols):
+    def _generate_features(self, df: pd.DataFrame, new_feat_cols: list) -> pd.DataFrame:
         """
         Generate additional features based on the feature formulas for all data points in the df.
         Only works after the model was fitted.
@@ -189,12 +186,12 @@ class AutoFeatModel(BaseEstimator):
         if not new_feat_cols[0] in self.feature_formulas_:
             raise RuntimeError("[AutoFeat] First call fit or fit_transform to generate the features!")
         if self.verbose:
-            print("[AutoFeat] Computing %i new features." % len(new_feat_cols))
+            print(f"[AutoFeat] Computing {len(new_feat_cols)} new features.")
         # generate all good feature; unscaled this time
         feat_array = np.zeros((len(df), len(new_feat_cols)))
         for i, expr in enumerate(new_feat_cols):
             if self.verbose:
-                print("[AutoFeat] %5i/%5i new features" % (i, len(new_feat_cols)), end="\r")
+                print(f"[AutoFeat] {i:5}/{len(new_feat_cols):5} new features", end="\r")
             if expr not in self.feature_functions_:
                 # generate a substitution expression based on all the original symbols of the original features
                 # for the given generated feature in good cols
@@ -210,7 +207,7 @@ class AutoFeatModel(BaseEstimator):
                         f = lambdify([self.feature_formulas_[c] for c in cols], self.feature_formulas_[expr])
                         f_jit = nb.njit(f)
                     except Exception:
-                        print("[AutoFeat] Error while processing expression: %r" % expr)
+                        print(f"[AutoFeat] Error while processing expression: {expr}")
                         raise
                 self.feature_functions_[expr] = (cols, f, f_jit)
             else:
@@ -230,15 +227,41 @@ class AutoFeatModel(BaseEstimator):
                     feat_array[not_na_idx, i] = feat
                     feat_array[~not_na_idx, i] = np.nan
                 except RuntimeWarning:
-                    print("[AutoFeat] WARNING: Problem while evaluating expression: %r with columns %r" % (expr, cols),
-                          " - is the data in a different range then when calling .fit()? Are maybe some values 0 that shouldn't be?")
+                    print(
+                        f"[AutoFeat] WARNING: Problem while evaluating expression: {expr} with columns {cols}",
+                        " - is the data in a different range then when calling .fit()? Are maybe some values 0 that shouldn't be?",
+                    )
                     raise
         if self.verbose:
-            print("[AutoFeat] %5i/%5i new features ...done." % (len(new_feat_cols), len(new_feat_cols)))
+            print(f"[AutoFeat] {len(new_feat_cols):5}/{len(new_feat_cols):5} new features ...done.")
         df = df.join(pd.DataFrame(feat_array, columns=new_feat_cols, index=df.index))
         return df
 
-    def fit_transform(self, X, y):
+    def _X2df(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
+        """
+        Helper function that ensures correctness of the input data for classification tasks.
+        Inputs:
+            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
+        Returns:
+            - df: transformed dataframe
+        """
+        # store column names as they'll be lost in the other check
+        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
+        # check input variables
+        X = check_array(X, dtype=None)
+        if not cols:
+            cols = [f"x{i:03}" for i in range(X.shape[1])]
+        # transform X into a dataframe (again)
+        df = pd.DataFrame(X, columns=cols)
+        # do we need to call transform?
+        if not list(df.columns) == self.all_columns_:
+            temp = self.always_return_numpy
+            self.always_return_numpy = False
+            df = self.transform(df)
+            self.always_return_numpy = temp
+        return df
+
+    def fit_transform(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Fits the regression model and returns a new dataframe with the additional features.
 
@@ -264,7 +287,7 @@ class AutoFeatModel(BaseEstimator):
             # the additional zeros in the name are because of the variable check in _generate_features,
             # where we check if the column name occurs in the the expression. this would lead to many
             # false positives if we have features x1 and x10...x19 instead of x001...x019.
-            cols = ["x%03i" % i for i in range(X.shape[1])]
+            cols = [f"x{i:03}" for i in range(X.shape[1])]
         self.original_columns_ = cols
         # transform X into a dataframe (again)
         df = pd.DataFrame(X, columns=cols)
@@ -275,7 +298,7 @@ class AutoFeatModel(BaseEstimator):
             fcols = []
             for c in self.feateng_cols:
                 if c not in self.original_columns_:
-                    raise ValueError("[AutoFeat] feateng_col %r not in df.columns" % c)
+                    raise ValueError(f"[AutoFeat] feateng_col {c} not in df.columns")
                 if c in self.categorical_cols_map_:
                     fcols.extend(self.categorical_cols_map_[c])
                 else:
@@ -294,12 +317,16 @@ class AutoFeatModel(BaseEstimator):
         n_cols = n_cols_generated(len(self.feateng_cols_), self.feateng_steps, len(self.transformations))
         n_gb = (len(df) * n_cols) / 250000000
         if self.verbose:
-            print("[AutoFeat] The %i step feature engineering process could generate up to %i features." % (self.feateng_steps, n_cols))
-            print("[AutoFeat] With %i data points this new feature matrix would use about %.2f gb of space." % (len(df), n_gb))
+            print(
+                f"[AutoFeat] The {self.feateng_steps} step feature engineering process could generate up to {n_cols} features."
+            )
+            print(f"[AutoFeat] With {len(df)} data points this new feature matrix would use about {n_gb:.2f} gb of space.")
         if self.max_gb and n_gb > self.max_gb:
             n_rows = int(self.max_gb * 250000000 / n_cols)
             if self.verbose:
-                print("[AutoFeat] As you specified a limit of %.1d gb, the number of data points is subsampled to %i" % (self.max_gb, n_rows))
+                print(
+                    f"[AutoFeat] As you specified a limit of {self.max_gb:.1f} gb, the number of data points is subsampled to {n_rows}"
+                )
             subsample_idx = np.random.permutation(list(df.index))[:n_rows]
             df_subs = df.iloc[subsample_idx]
             df_subs.reset_index(drop=True, inplace=True)
@@ -308,8 +335,14 @@ class AutoFeatModel(BaseEstimator):
             df_subs = df.copy()
             target_sub = target.copy()
         # generate features
-        df_subs, self.feature_formulas_ = engineer_features(df_subs, self.feateng_cols_, _parse_units(self.units, verbose=self.verbose),
-                                                            self.feateng_steps, self.transformations, self.verbose)
+        df_subs, self.feature_formulas_ = engineer_features(
+            df_subs,
+            self.feateng_cols_,
+            _parse_units(self.units, verbose=self.verbose),
+            self.feateng_steps,
+            self.transformations,
+            self.verbose,
+        )
         # select predictive features
         if self.featsel_runs <= 0:
             if self.verbose:
@@ -317,12 +350,14 @@ class AutoFeatModel(BaseEstimator):
             good_cols = df_subs.columns
         else:
             if self.problem_type in ("regression", "classification"):
-                good_cols = select_features(df_subs, target_sub, self.featsel_runs, None, self.problem_type, self.n_jobs, self.verbose)
+                good_cols = select_features(
+                    df_subs, target_sub, self.featsel_runs, None, self.problem_type, self.n_jobs, self.verbose
+                )
                 # if no features were selected, take the original features
                 if not good_cols:
                     good_cols = list(df.columns)
             else:
-                print("[AutoFeat] WARNING: Unknown problem_type %r - not performing feature selection." % self.problem_type)
+                print(f"[AutoFeat] WARNING: Unknown problem_type {self.problem_type} - not performing feature selection.")
                 good_cols = df_subs.columns
         # filter out those columns that were original features or generated otherwise
         self.new_feat_cols_ = [c for c in good_cols if c not in list(df.columns)]
@@ -330,6 +365,8 @@ class AutoFeatModel(BaseEstimator):
         # re-generate all good feature again; for all data points this time
         self.feature_functions_ = {}
         df = self._generate_features(df, self.new_feat_cols_)
+        # to prevent an error because sometimes the column names are numpy.str_ instead of normal str
+        df.columns = [str(c) for c in df.columns]
         # filter out unnecessary junk from self.feature_formulas_
         self.feature_formulas_ = {f: self.feature_formulas_[f] for f in self.new_feat_cols_ + self.feateng_cols_}
         self.feature_functions_ = {f: self.feature_functions_[f] for f in self.new_feat_cols_}
@@ -337,7 +374,9 @@ class AutoFeatModel(BaseEstimator):
         # train final prediction model on all selected features
         if self.verbose:
             # final dataframe contains original columns and good additional columns
-            print("[AutoFeat] Final dataframe with %i feature columns (%i new)." % (len(df.columns), len(df.columns) - len(self.original_columns_)))
+            print(
+                f"[AutoFeat] Final dataframe with {len(df.columns)} feature columns ({len(df.columns) - len(self.original_columns_)} new)."
+            )
 
         # train final prediction model
         if self.problem_type == "regression":
@@ -345,11 +384,11 @@ class AutoFeatModel(BaseEstimator):
         elif self.problem_type == "classification":
             model = lm.LogisticRegressionCV(cv=5, class_weight="balanced")
         else:
-            print("[AutoFeat] WARNING: Unknown problem_type %r - not fitting a prediction model." % self.problem_type)
+            print(f"[AutoFeat] WARNING: Unknown problem_type {self.problem_type} - not fitting a prediction model.")
             model = None
         if model is not None:
             if self.verbose:
-                print("[AutoFeat] Training final %s model." % self.problem_type)
+                print(f"[AutoFeat] Training final {self.problem_type} model.")
             X = df[self.good_cols_].to_numpy()
             with warnings.catch_warnings():
                 warnings.simplefilter("ignore")
@@ -370,20 +409,20 @@ class AutoFeatModel(BaseEstimator):
                 for c in sorted(weights, key=lambda x: abs(weights[x]), reverse=True):
                     if abs(weights[c]) < 1e-5:
                         break
-                    print("%.6f * %s" % (weights[c], c))
-                print("[AutoFeat] Final score: %.4f" % model.score(X, target))
+                    print(f"{weights[c]:.6f} * {c}")
+                print(f"[AutoFeat] Final score: {model.score(X, target):.4f}")
         if self.always_return_numpy:
             return df.to_numpy()
         return df
 
-    def fit(self, X, y):
+    def fit(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame):
         if self.verbose:
             print("[AutoFeat] Warning: This just calls fit_transform() but does not return the transformed dataframe.")
             print("[AutoFeat] It is much more efficient to call fit_transform() instead of fit() and transform()!")
         _ = self.fit_transform(X, y)  # noqa
         return self
 
-    def transform(self, X):
+    def transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Inputs:
             - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
@@ -398,7 +437,7 @@ class AutoFeatModel(BaseEstimator):
         # check input variables
         X = check_array(X, force_all_finite="allow-nan", dtype=None)
         if not cols:
-            cols = ["x%03i" % i for i in range(X.shape[1])]
+            cols = [f"x{i:03}" for i in range(X.shape[1])]
         if not cols == self.original_columns_:
             raise ValueError("[AutoFeat] Not the same features as when calling fit.")
         # transform X into a dataframe (again)
@@ -409,35 +448,24 @@ class AutoFeatModel(BaseEstimator):
         df = self._apply_pi_theorem(df)
         # generate engineered features
         df = self._generate_features(df, self.new_feat_cols_)
+        # to prevent an error because sometimes the column names are numpy.str_ instead of normal str
+        df.columns = [str(c) for c in df.columns]
         if self.always_return_numpy:
             return df.to_numpy()
         return df
 
-    def predict(self, X):
+    def predict(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Inputs:
             - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
         Returns:
-            - y_pred: predicted targets return by prediction_model.predict()
+            - y_pred: predicted targets returned by prediction_model.predict()
         """
         check_is_fitted(self, ["prediction_model_"])
-        # store column names as they'll be lost in the other check
-        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
-        # check input variables
-        X = check_array(X, dtype=None)
-        if not cols:
-            cols = ["x%03i" % i for i in range(X.shape[1])]
-        # transform X into a dataframe (again)
-        df = pd.DataFrame(X, columns=cols)
-        # do we need to call transform?
-        if not list(df.columns) == self.all_columns_:
-            temp = self.always_return_numpy
-            self.always_return_numpy = False
-            df = self.transform(df)
-            self.always_return_numpy = temp
+        df = self._X2df(X)
         return self.prediction_model_.predict(df[self.good_cols_].to_numpy())
 
-    def score(self, X, y):
+    def score(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Inputs:
             - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
@@ -451,7 +479,7 @@ class AutoFeatModel(BaseEstimator):
         # check input variables
         X, target = check_X_y(X, y, y_numeric=self.problem_type == "regression", dtype=None)
         if not cols:
-            cols = ["x%03i" % i for i in range(X.shape[1])]
+            cols = [f"x{i:03}" for i in range(X.shape[1])]
         # transform X into a dataframe (again)
         df = pd.DataFrame(X, columns=cols)
         # do we need to call transform?
@@ -472,20 +500,32 @@ class AutoFeatRegressor(AutoFeatModel, BaseEstimator, RegressorMixin):
 
     def __init__(
         self,
-        categorical_cols=None,
-        feateng_cols=None,
-        units=None,
-        feateng_steps=2,
-        featsel_runs=5,
-        max_gb=None,
-        transformations=("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
-        apply_pi_theorem=True,
-        always_return_numpy=False,
-        n_jobs=1,
-        verbose=0,
+        categorical_cols: list | None = None,
+        feateng_cols: list | None = None,
+        units: dict | None = None,
+        feateng_steps: int = 2,
+        featsel_runs: int = 5,
+        max_gb: int | None = None,
+        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
+        apply_pi_theorem: bool = True,
+        always_return_numpy: bool = False,
+        n_jobs: int = 1,
+        verbose: int = 0,
     ):
-        super().__init__("regression", categorical_cols, feateng_cols, units, feateng_steps,
-                         featsel_runs, max_gb, transformations, apply_pi_theorem, always_return_numpy, n_jobs, verbose)
+        super().__init__(
+            "regression",
+            categorical_cols,
+            feateng_cols,
+            units,
+            feateng_steps,
+            featsel_runs,
+            max_gb,
+            transformations,
+            apply_pi_theorem,
+            always_return_numpy,
+            n_jobs,
+            verbose,
+        )
 
 
 class AutoFeatClassifier(AutoFeatModel, BaseEstimator, ClassifierMixin):
@@ -493,17 +533,40 @@ class AutoFeatClassifier(AutoFeatModel, BaseEstimator, ClassifierMixin):
 
     def __init__(
         self,
-        categorical_cols=None,
-        feateng_cols=None,
-        units=None,
-        feateng_steps=2,
-        featsel_runs=5,
-        max_gb=None,
-        transformations=("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
-        apply_pi_theorem=True,
-        always_return_numpy=False,
-        n_jobs=1,
-        verbose=0,
+        categorical_cols: list | None = None,
+        feateng_cols: list | None = None,
+        units: dict | None = None,
+        feateng_steps: int = 2,
+        featsel_runs: int = 5,
+        max_gb: int | None = None,
+        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
+        apply_pi_theorem: bool = True,
+        always_return_numpy: bool = False,
+        n_jobs: int = 1,
+        verbose: int = 0,
     ):
-        super().__init__("classification", categorical_cols, feateng_cols, units, feateng_steps,
-                         featsel_runs, max_gb, transformations, apply_pi_theorem, always_return_numpy, n_jobs, verbose)
+        super().__init__(
+            "classification",
+            categorical_cols,
+            feateng_cols,
+            units,
+            feateng_steps,
+            featsel_runs,
+            max_gb,
+            transformations,
+            apply_pi_theorem,
+            always_return_numpy,
+            n_jobs,
+            verbose,
+        )
+
+    def predict_proba(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
+        """
+        Inputs:
+            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
+        Returns:
+            - y_pred: predicted targets probabilities returned by prediction_model.predict_proba()
+        """
+        check_is_fitted(self, ["prediction_model_"])
+        df = self._X2df(X)
+        return self.prediction_model_.predict_proba(df[self.good_cols_].to_numpy())
