@@ -1,9 +1,7 @@
-# -*- coding: utf-8 -*-
 # Author: Franziska Horn <cod3licious@gmail.com>
 # License: MIT
 
-from __future__ import unicode_literals, division, print_function, absolute_import
-from builtins import zip
+from __future__ import annotations
 import warnings
 import numpy as np
 import pandas as pd
@@ -15,7 +13,7 @@ import sklearn.linear_model as lm
 from .nb_utils import nb_standard_scale
 
 
-def _add_noise_features(X):
+def _add_noise_features(X: np.ndarray):
     """
     Adds 3-1.5*d additional noise features to X.
 
@@ -30,12 +28,14 @@ def _add_noise_features(X):
         rand_noise = nb_standard_scale(np.random.permutation(X.flatten()).reshape(X.shape))
         X = np.hstack([X, rand_noise])
     # normally distributed noise
-    rand_noise = np.random.randn(X.shape[0], max(3, int(0.5*n_feat)))
+    rand_noise = np.random.randn(X.shape[0], max(3, int(0.5 * n_feat)))
     X = np.hstack([X, rand_noise])
     return X
 
 
-def _noise_filtering(X, target, good_cols=[], problem_type="regression"):
+def _noise_filtering(
+    X: np.ndarray, target: np.ndarray, good_cols: list | None = None, problem_type: str = "regression"
+) -> list:
     """
     Trains a prediction model with additional noise features and selects only those of the
     original features that have a higher coefficient than any of the noise features.
@@ -49,16 +49,16 @@ def _noise_filtering(X, target, good_cols=[], problem_type="regression"):
         - good_cols: list of noise filtered column names
     """
     n_feat = X.shape[1]
-    assert len(good_cols) == n_feat, "fewer column names provided than features in X."
-    if not good_cols:
+    if good_cols is None or not len(good_cols):
         good_cols = list(range(n_feat))
+    assert len(good_cols) == n_feat, "fewer column names provided than features in X."
     # perform noise filtering on these features
     if problem_type == "regression":
         model = lm.LassoLarsCV(cv=5, eps=1e-8)
     elif problem_type == "classification":
         model = lm.LogisticRegressionCV(cv=5, penalty="l1", solver="saga", class_weight="balanced")
     else:
-        print("[featsel] WARNING: Unknown problem_type %r - not performing noise filtering." % problem_type)
+        print(f"[featsel] WARNING: Unknown problem_type {problem_type} - not performing noise filtering.")
         model = None
     if model is not None:
         X = _add_noise_features(X)
@@ -76,14 +76,14 @@ def _noise_filtering(X, target, good_cols=[], problem_type="regression"):
         else:
             # model.coefs_ is n_classes x n_features, but we need n_features
             coefs = np.max(np.abs(model.coef_), axis=0)
-        weights = dict(zip(good_cols, coefs[:len(good_cols)]))
+        weights = dict(zip(good_cols, coefs[: len(good_cols)]))
         # only include features that are more important than our known noise features
         noise_w_thr = np.max(coefs[n_feat:])
         good_cols = [c for c in good_cols if weights[c] > noise_w_thr]
     return good_cols
 
 
-def _select_features_1run(df, target, problem_type="regression", verbose=0):
+def _select_features_1run(df: pd.DataFrame, target: np.ndarray, problem_type: str = "regression", verbose: int = 0) -> list:
     """
     One feature selection run.
 
@@ -97,14 +97,14 @@ def _select_features_1run(df, target, problem_type="regression", verbose=0):
         - good_cols: list of column names for df with which a prediction model can be trained
     """
     if df.shape[0] <= 1:
-        raise ValueError("n_samples = {}".format(df.shape[0]))
+        raise ValueError(f"n_samples = {df.shape[0]}")
     # initial selection of too few but (hopefully) relevant features
     if problem_type == "regression":
         model = lm.LassoLarsCV(cv=5, eps=1e-8)
     elif problem_type == "classification":
         model = lm.LogisticRegressionCV(cv=5, penalty="l1", solver="saga", class_weight="balanced")
     else:
-        print("[featsel] WARNING: Unknown problem_type %r - not performing feature selection!" % problem_type)
+        print(f"[featsel] WARNING: Unknown problem_type {problem_type} - not performing feature selection!")
         return []
     with warnings.catch_warnings():
         warnings.simplefilter("ignore")
@@ -122,22 +122,22 @@ def _select_features_1run(df, target, problem_type="regression", verbose=0):
         # model.coefs_ is n_classes x n_features, but we need n_features
         coefs = np.max(np.abs(model.coef_), axis=0)
     # weight threshold: select at most 0.2*n_train initial features
-    thr = sorted(coefs, reverse=True)[min(df.shape[1]-1, df.shape[0]//5)]
+    thr = sorted(coefs, reverse=True)[min(df.shape[1] - 1, df.shape[0] // 5)]
     initial_cols = list(df.columns[coefs > thr])
     # noise filter
     initial_cols = _noise_filtering(df[initial_cols].to_numpy(), target, initial_cols, problem_type)
-    good_cols = set(initial_cols)
+    good_cols_set = set(initial_cols)
     if verbose > 0:
-        print("[featsel]\t %i initial features." % len(initial_cols))
+        print(f"[featsel]\t {len(initial_cols)} initial features.")
     # add noise features
     X_w_noise = _add_noise_features(df[initial_cols].to_numpy())
     # go through all remaining features in splits of n_feat <= 0.5*n_train
     other_cols = list(np.random.permutation(list(set(df.columns).difference(initial_cols))))
     if other_cols:
-        n_splits = int(np.ceil(len(other_cols)/max(10, 0.5*df.shape[0]-len(initial_cols))))
-        split_size = int(np.ceil(len(other_cols)/n_splits))
+        n_splits = int(np.ceil(len(other_cols) / max(10, 0.5 * df.shape[0] - len(initial_cols))))
+        split_size = int(np.ceil(len(other_cols) / n_splits))
         for i in range(n_splits):
-            current_cols = other_cols[i*split_size:min(len(other_cols), (i+1)*split_size)]
+            current_cols = other_cols[i * split_size : min(len(other_cols), (i + 1) * split_size)]
             X = np.hstack([df[current_cols].to_numpy(), X_w_noise])
             if problem_type == "regression":
                 model = lm.LassoLarsCV(cv=5, eps=1e-8)
@@ -158,21 +158,31 @@ def _select_features_1run(df, target, problem_type="regression", verbose=0):
             else:
                 # model.coefs_ is n_classes x n_features, but we need n_features
                 coefs = np.max(np.abs(model.coef_), axis=0)
-            weights = dict(zip(current_cols, coefs[:len(current_cols)]))
+            weights = dict(zip(current_cols, coefs[: len(current_cols)]))
             # only include features that are more important than our known noise features
-            noise_w_thr = np.max(coefs[len(current_cols):])
-            good_cols.update([c for c in weights if abs(weights[c]) > noise_w_thr])
+            noise_w_thr = np.max(coefs[len(current_cols) :])
+            good_cols_set.update([c for c in weights if abs(weights[c]) > noise_w_thr])
             if verbose > 0:
-                print("[featsel]\t Split %2i/%i: %3i candidate features identified." % (i+1, n_splits, len(good_cols)), end="\r")
+                print(
+                    f"[featsel]\t Split {i + 1:2}/{n_splits}: {len(good_cols_set):3} candidate features identified.", end="\r"
+                )
     # noise filtering on the combination of features
-    good_cols = list(good_cols)
+    good_cols = list(good_cols_set)
     good_cols = _noise_filtering(df[good_cols].to_numpy(), target, good_cols, problem_type)
     if verbose > 0:
-        print("\n[featsel]\t Selected %3i features after noise filtering." % len(good_cols))
+        print(f"\n[featsel]\t Selected {len(good_cols):3} features after noise filtering.")
     return good_cols
 
 
-def select_features(df, target, featsel_runs=5, keep=None, problem_type="regression", n_jobs=1, verbose=0):
+def select_features(
+    df: pd.DataFrame,
+    target: np.ndarray,
+    featsel_runs: int = 5,
+    keep: list | None = None,
+    problem_type: str = "regression",
+    n_jobs: int = 1,
+    verbose: int = 0,
+) -> list:
     """
     Selects predictive features given the data and targets.
 
@@ -213,12 +223,12 @@ def select_features(df, target, featsel_runs=5, keep=None, problem_type="regress
 
     # select good features in k runs in parallel
     # by doing sort of a cross-validation (i.e., randomly subsample data points)
-    def run_select_features(i):
+    def run_select_features(i: int):
         if verbose > 0:
-            print("[featsel] Feature selection run %i/%i" % (i+1, featsel_runs))
+            print(f"[featsel] Feature selection run {i + 1}/{featsel_runs}")
         np.random.seed(i)
-        rand_idx = np.random.permutation(df_scaled.index)[:max(10, int(0.85 * len(df_scaled)))]
-        return _select_features_1run(df_scaled.iloc[rand_idx], target_scaled[rand_idx], problem_type, verbose=verbose-1)
+        rand_idx = np.random.permutation(df_scaled.index)[: max(10, int(0.85 * len(df_scaled)))]
+        return _select_features_1run(df_scaled.iloc[rand_idx], target_scaled[rand_idx], problem_type, verbose=verbose - 1)
 
     if featsel_runs >= 1 and problem_type in ("regression", "classification"):
         if n_jobs == 1 or featsel_runs == 1:
@@ -227,17 +237,22 @@ def select_features(df, target, featsel_runs=5, keep=None, problem_type="regress
             for i in range(featsel_runs):
                 selected_columns.extend(run_select_features(i))
         else:
-            def flatten_lists(l):
+
+            def flatten_lists(l: list):
                 return [item for sublist in l for item in sublist]
 
-            selected_columns = flatten_lists(Parallel(n_jobs=n_jobs, verbose=100*verbose)(delayed(run_select_features)(i) for i in range(featsel_runs)))
+            selected_columns = flatten_lists(
+                Parallel(n_jobs=n_jobs, verbose=100 * verbose)(delayed(run_select_features)(i) for i in range(featsel_runs))
+            )
 
         if selected_columns:
             selected_columns = Counter(selected_columns)
             # sort by frequency, but down weight longer formulas to break ties
-            selected_columns = sorted(selected_columns, key=lambda x: selected_columns[x] - 0.000001*len(str(x)), reverse=True)
+            selected_columns = sorted(
+                selected_columns, key=lambda x: selected_columns[x] - 0.000001 * len(str(x)), reverse=True
+            )
             if verbose > 0:
-                print("[featsel] %i features after %i feature selection runs" % (len(selected_columns), featsel_runs))
+                print(f"[featsel] {len(selected_columns)} features after {featsel_runs} feature selection runs")
             # correlation filtering
             selected_columns = keep + [c for c in selected_columns if c not in keep]
             if not keep:
@@ -253,31 +268,30 @@ def select_features(df, target, featsel_runs=5, keep=None, problem_type="regress
                     if np.max(np.abs(correlations[c].ravel()[:i])) < 0.9:
                         good_cols.append(c)
             if verbose > 0:
-                print("[featsel] %i features after correlation filtering" % len(good_cols))
+                print(f"[featsel] {len(good_cols)} features after correlation filtering")
 
     # perform noise filtering on these features
     good_cols = _noise_filtering(df_scaled[good_cols].to_numpy(), target_scaled, good_cols, problem_type)
     if verbose > 0:
-        print("[featsel] %i features after noise filtering" % len(good_cols))
+        print(f"[featsel] {len(good_cols)} features after noise filtering")
         if not good_cols:
             print("[featsel] WARNING: Not a single good features was found...")
 
     # add keep columns back in
     good_cols = keep + [c for c in good_cols if c not in keep]
     if verbose > 0 and keep:
-        print("[featsel] %i final features selected (including %i original keep features)." % (len(good_cols), len(keep)))
+        print(f"[featsel] {len(good_cols)} final features selected (including {len(keep)} original keep features).")
     return good_cols
 
 
 class FeatureSelector(BaseEstimator):
-
     def __init__(
         self,
-        problem_type="regression",
-        featsel_runs=5,
-        keep=None,
-        n_jobs=1,
-        verbose=0,
+        problem_type: str = "regression",
+        featsel_runs: int = 5,
+        keep: list | None = None,
+        n_jobs: int = 1,
+        verbose: int = 0,
     ):
         """
         multi-step cross-validated feature selection
@@ -301,7 +315,7 @@ class FeatureSelector(BaseEstimator):
         self.n_jobs = n_jobs
         self.verbose = verbose
 
-    def fit(self, X, y):
+    def fit(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame):
         """
         Selects features.
 
@@ -317,16 +331,18 @@ class FeatureSelector(BaseEstimator):
         # check input variables
         X, target = check_X_y(X, y, y_numeric=self.problem_type == "regression")
         if not cols:
-            cols = ["x%i" % i for i in range(X.shape[1])]
+            cols = [f"x{i}" for i in range(X.shape[1])]
         self.original_columns_ = cols
         # transform X into a dataframe (again)
         df = pd.DataFrame(X, columns=cols)
         # do the feature selection
-        self.good_cols_ = select_features(df, target, self.featsel_runs, self.keep, self.problem_type, self.n_jobs, self.verbose)
+        self.good_cols_ = select_features(
+            df, target, self.featsel_runs, self.keep, self.problem_type, self.n_jobs, self.verbose
+        )
         self.n_features_in_ = X.shape[1]
         return self
 
-    def transform(self, X):
+    def transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Inputs:
             - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
@@ -345,7 +361,7 @@ class FeatureSelector(BaseEstimator):
         # check input variables
         X = check_array(X, force_all_finite="allow-nan")
         if not cols:
-            cols = ["x%i" % i for i in range(X.shape[1])]
+            cols = [f"x{i}" for i in range(X.shape[1])]
         if not cols == self.original_columns_:
             raise ValueError("[FeatureSelector] Not the same features as when calling fit.")
         # transform X into a dataframe (again) and select columns
@@ -355,7 +371,7 @@ class FeatureSelector(BaseEstimator):
             new_X = new_X.to_numpy()
         return new_X
 
-    def fit_transform(self, X, y):
+    def fit_transform(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
         """
         Selects features and returns only those selected.
 
