@@ -1,12 +1,12 @@
-# -*- coding: utf-8 -*-
 # Author: Jeethu Rao <jboloor@acm.org>
 # License: MIT
 
+from __future__ import annotations
 import numba as nb
 import numpy as np
 
 
-@nb.njit(inline='always')
+@nb.njit(inline="always")
 def nb_apply_along_axis(func1d, axis, arr):
     assert arr.ndim == 2
     assert axis in [0, 1]
